import Graph from 'graphology';
import { circular, random } from 'graphology-layout';
import forceAtlas2, {
  ForceAtlas2Settings,
} from 'graphology-layout-forceatlas2';
import noverlap from 'graphology-layout-noverlap';
import { RandomLayoutOptions } from 'graphology-layout/random';
import { NoverlapSettings } from 'graphology-library/layout-noverlap';
import { Attributes } from 'graphology-types';
import { Layout } from './Layout';
import { ILayoutFactory, LayoutAlgorithm } from './layout-creator-usecase';

export type GraphologyProvider = 'Graphology';

export type GraphologyLayoutAlgorithms =
  | `${GraphologyProvider}_circular`
  | `${GraphologyProvider}_random`
  | `${GraphologyProvider}_noverlap`
  | `${GraphologyProvider}_forceAtlas2`;

/**
 * This is a ConcreteCreator
 */
export class GraphologyFactory
  implements ILayoutFactory<GraphologyLayoutAlgorithms>
{
  createLayout(layoutAlgorithm: GraphologyLayoutAlgorithms): Graphology | null {
    switch (layoutAlgorithm) {
      case 'Graphology_random':
        return new GraphologyRandom();
      case 'Graphology_circular':
        return new GraphologyCircular();
      case 'Graphology_noverlap':
        return new GraphologyNoverlap();
      case 'Graphology_forceAtlas2':
        return new GraphologyForceAtlas2();
      default:
        return null;
    }
  }
}

export abstract class Graphology extends Layout<GraphologyProvider> {
  height = 100;
  width = 100;
  constructor(public override algorithm: LayoutAlgorithm<GraphologyProvider>) {
    super('Graphology', algorithm);
    this.setDimensions(100, 200);
  }

  public specialGraphologyFunction() {
    // graph.forEachNode((node, attr) => {
    //   graph.setNodeAttribute(node, 'x', 0);
    //   graph.setNodeAttribute(node, 'y', 0);
    // });
  }

  public setDimensions(width = 100, height = 100) {
    this.width = width;
    this.height = height;
  }
}

/**
 * This is a ConcreteProduct
 */
export class GraphologyCircular extends Graphology {
  constructor() {
    super('Graphology_circular');
  }

  public override layout(
    graph: Graph<Attributes, Attributes, Attributes>
  ): void {
    // To directly assign the positions to the nodes:
    circular.assign(graph, {
      scale: 100,
    });
  }
}

const DEFAULT_RANDOM_SETTINGS: RandomLayoutOptions = {
  scale: 250,
};
/**
 * This is a ConcreteProduct
 */
export class GraphologyRandom extends Graphology {
  constructor() {
    super('Graphology_random');
  }

  public override layout(
    graph: Graph<Attributes, Attributes, Attributes>
  ): void {
    // const positions = random(graph);

    // To directly assign the positions to the nodes:
    random.assign(graph, DEFAULT_RANDOM_SETTINGS);
  }
}

const DEFAULT_NOVERLAP_SETTINGS: NoverlapSettings = {
  margin: 40,
  ratio: 40,
  // gridSize: 50,

  // gridSize ?number 20: number of grid cells horizontally and vertically subdivising the graph’s space. This is used as an optimization scheme. Set it to 1 and you will have O(n²) time complexity, which can sometimes perform better with very few nodes.
  // margin ?number 5: margin to keep between nodes.
  // expansion ?number 1.1: percentage of current space that nodes could attempt to move outside of.
  // ratio ?number 1.0: ratio scaling node sizes.
  // speed ?number 3: dampening factor that will slow down node movements to ease the overall process.
};

/**
 * This is a ConcreteProduct
 */
export class GraphologyNoverlap extends Graphology {
  constructor() {
    super('Graphology_noverlap');
  }

  public override layout(
    graph: Graph<Attributes, Attributes, Attributes>
  ): void {
    // To directly assign the positions to the nodes:
    noverlap.assign(graph, {
      maxIterations: 5000,
      settings: DEFAULT_NOVERLAP_SETTINGS,
    });
  }
}

const DEFAULT_FORCEATLAS2_SETTINGS: ForceAtlas2Settings = {
  gravity: 10,
  adjustSizes: true,
  linLogMode: true,

  // adjustSizes ?boolean false: should the node’s sizes be taken into account?
  // barnesHutOptimize ?boolean false: whether to use the Barnes-Hut approximation to compute repulsion in O(n*log(n)) rather than default O(n^2), n being the number of nodes.
  // barnesHutTheta ?number 0.5: Barnes-Hut approximation theta parameter.
  // edgeWeightInfluence ?number 1: influence of the edge’s weights on the layout. To consider edge weight, don’t forget to pass weighted as true when applying the synchronous layout or when instantiating the worker.
  // gravity ?number 1: strength of the layout’s gravity.
  // linLogMode ?boolean false: whether to use Noack’s LinLog model.
  // outboundAttractionDistribution ?boolean false
  // scalingRatio ?number 1
  // slowDown ?number 1
  // strongGravityMode ?boolean false
};

/**
 * This is a ConcreteProduct
 */
export class GraphologyForceAtlas2 extends Graphology {
  constructor() {
    super('Graphology_forceAtlas2');
  }

  public override layout(
    graph: Graph<Attributes, Attributes, Attributes>
  ): void {
    // To directly assign the positions to the nodes:

    forceAtlas2.assign(graph, {
      iterations: 500000,
      settings: DEFAULT_FORCEATLAS2_SETTINGS,
    });
  }
}
